@@ -343,8 +343,6 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                                      "file_toolbar"])
 
         self.widget.ensure_style()
-        self.on_style_updated(self.widget)
-        self.widget.connect("style-updated", self.on_style_updated)
 
         self.custom_labels = []
         self.set_num_panes(num_panes)
@@ -378,7 +376,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             column.pack_start(renicon, False)
             column.pack_start(rentext, True)
             column.set_attributes(rentext, markup=col_index(tree.COL_TEXT, i),
-                                  foreground=col_index(tree.COL_FG, i),
+                                  foreground_rgba=col_index(tree.COL_FG, i),
                                   style=col_index(tree.COL_STYLE, i),
                                   weight=col_index(tree.COL_WEIGHT, i),
                                   strikethrough=col_index(tree.COL_STRIKE, i))
@@ -445,29 +443,6 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
         self._scan_in_progress = 0
 
-    def on_style_updated(self, widget):
-        style = widget.get_style_context()
-
-        def lookup(name, default):
-            found, colour = style.lookup_color(name)
-            if not found:
-                colour = Gdk.RGBA()
-                colour.parse(default)
-            return colour
-
-        self.fill_colors = {"insert"  : lookup("insert-bg", "DarkSeaGreen1"),
-                            "delete"  : lookup("delete-bg", "White"),
-                            "replace" : lookup("replace-bg", "#ddeeff"),
-                            "error"   : lookup("error-bg", "#fce94f")}
-        self.line_colors = {"insert"  : lookup("insert-outline", "#77f077"),
-                            "delete"  : lookup("delete-outline", "Grey"),
-                            "replace" : lookup("replace-outline", "#8bbff3"),
-                            "error"   : lookup("error-outline", "#edd400")}
-
-        for diffmap in self.diffmap:
-            diffmap.set_color_scheme([self.fill_colors, self.line_colors])
-        self.queue_draw()
-
     def queue_draw(self):
         for treeview in self.treeview:
             treeview.queue_draw()
@@ -1487,7 +1462,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             for (w, i) in zip(self.diffmap, (0, n - 1)):
                 scroll = self.scrolledwindow[i].get_vscrollbar()
                 idx = 1 if i else 0
-                w.setup(scroll, self.get_state_traversal(idx), [self.fill_colors, self.line_colors])
+                w.setup(scroll, self.get_state_traversal(idx))
 
             for w in self.linkmap:
                 w.associate(self)
