@@ -762,21 +762,39 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                                                 None, True)
             self.queue_draw()
 
-    def _filter_text(self, txt):
-        def killit(m):
-            assert m.group().count("\n") == 0
-            if len(m.groups()):
-                s = m.group()
-                for g in m.groups():
-                    if g:
-                        s = s.replace(g,"")
-                return s
-            else:
-                return ""
+    def _filter_text(self, txt, buf, txt_start_iter, txt_end_iter):
+        dimmed_tag = buf.get_tag_table().lookup("dimmed")
+        buf.remove_tag(dimmed_tag, txt_start_iter, txt_end_iter)
+
         try:
-            for filt in self.text_filters:
-                if filt.active:
-                    txt = filt.filter.sub(killit, txt)
+            filter_ranges = []
+
+            active_filters = [f for f in self.text_filters if f.active]
+            for filt in active_filters:
+                for match in filt.filter.finditer(txt):
+                    # If there are no groups in the match, use the whole match
+                    if not filt.filter.groups:
+                        filter_ranges.append(match.span())
+                        continue
+
+                    # If there are groups in the regex, include all groups that
+                    # participated in the match
+                    for i in range(filt.filter.groups):
+                        span = match.span(i + 1)
+                        if span != (-1, -1):
+                            filter_ranges.append(span)
+
+            filter_ranges = misc.merge_intervals(filter_ranges)
+
+            for (start, end) in reversed(filter_ranges):
+                assert txt[start:end].count("\n") == 0
+                txt = txt[:start] + txt[end:]
+                start_iter = txt_start_iter.copy()
+                start_iter.forward_chars(start)
+                end_iter = txt_start_iter.copy()
+                end_iter.forward_chars(end)
+                buf.apply_tag(dimmed_tag, start_iter, end_iter)
+
         except AssertionError:
             if not self.warned_bad_comparison:
                 misc.error_dialog(
@@ -787,6 +805,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                         u"not be accurate.") % filt.label,
                 )
                 self.warned_bad_comparison = True
+
         return txt
 
     def after_text_insert_text(self, buf, it, newtext, textlen):
@@ -850,7 +869,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                     messagetype=Gtk.MessageType.QUESTION)
 
                 if resolve_response == Gtk.ResponseType.OK:
-                    conflict_file = self.textbuffer[1].data.filename
+                    bufdata = self.textbuffer[1].data
+                    conflict_file = bufdata.savefile or bufdata.filename
                     parent.command('resolve', [conflict_file])
 
         return response
@@ -1587,7 +1607,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                             text = text.replace('\n', k)
                         break
 
-        encoding = bufdata.encoding
+        encoding = bufdata.encoding or 'utf-8'
         while isinstance(text, unicode):
             try:
                 text = text.encode(encoding)
@@ -1964,7 +1984,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
     def clear_sync_points(self, action):
         self.syncpoints = []
         self.linediffer.syncpoints = []
-        for t in enumerate(self.textview):
+        for t in self.textview:
             t.syncpoints = []
         for mgr in self.msgarea_mgr:
             if mgr.get_msg_id() == FileDiff.MSG_SYNCPOINTS:
